@@ -98,7 +98,9 @@ def submit_job(
             Determines what to do when files already exist. Possible options include
             `error`, `warn_skip`, `silent_skip`, `overwrite`, and `silent_overwrite`.
         **kwargs:
-            Extra arguments are forwarded as template variables to the script
+            Extra arguments are forwarded as template variables to the script. A typical
+            use case is specifying the partition in a `qsub` script, which can be done
+            by giving the `partition` parameter here
 
     Returns:
         tuple: The result `(stdout, stderr)` of the submission call
@@ -119,6 +121,7 @@ def submit_job(
     ensure_directory_exists(log_folder)
 
     script_args = {
+        "PACKAGE_PATH": Path(__file__).parents[2],
         "LOG_FOLDER": log_folder,
         "JOB_NAME": name,
         "MODEL_FILE": escape_string(script),
@@ -157,6 +160,9 @@ def submit_job(
             else:
                 raise NotImplementedError(f"Unknown strategy `{overwrite_strategy}`")
 
+            # delete old output
+            output.unlink(missing_ok=True)
+
         # check whether output points to a directory or whether this should be a file
         if output.is_dir():
             script_args["OUTPUT_FOLDER"] = pipes.quote(str(output))
@@ -205,9 +211,10 @@ def submit_jobs(
     name_base: str = "job",
     parameters: Union[str, Dict[str, Any], None] = None,
     *,
+    output_format: str = "hdf",
     list_params: Optional[Iterable[str]] = None,
     **kwargs,
-) -> None:
+) -> int:
     """submit many jobs of the same script with different parameters to the cluster
 
     Args:
@@ -222,11 +229,16 @@ def submit_jobs(
             containing a JSON-encoded dictionary. All combinations of parameter values
             that are iterable and not strings and not part of `keep_list` are submitted
             as separate jobs.
+        output_format (str):
+            File extension determining the output format
         list_params (list):
             List of parameters that are meant to be lists. They will be submitted as
             individual parameters and not iterated over to produce multiple jobs.
         **kwargs:
             All additional parameters are forwarded to :func:`submit_job`.
+
+    Returns:
+        int: The number of jobs that have been submitted
     """
     if parameters is None:
         parameter_dict = {}
@@ -255,9 +267,14 @@ def submit_jobs(
         for values in itertools.product(*p_vary.values())
     ]
 
+    if not output_format.startswith("."):
+        output_format = "." + output_format
+
     # submit jobs with all parameter variations
     for p_job in tqdm(p_vary_list):
         params.update(p_job)
         name = get_job_name(name_base, p_job)
-        output = Path(output_folder) / f"{name}.hdf5"
+        output = Path(output_folder) / f"{name}{output_format}"
         submit_job(script, output=output, name=name, parameters=params, **kwargs)
+
+    return len(p_vary_list)
